/*
 * All of the documentation and software included in the
 * BLM Software is copyrighted by Jesse Davis and Pedro Domingos.
 * The arguments.cpp, arguments.h, hasharray.h, hash.h, hashint.h
 * flbfgsb.h, lbfgsb.h, random.h and timer.h come from the
 * Alchemy Software system. These files are denoted as coming from
 * Alchemy.
 * 
 * Copyright [2010] Jesse Davis and Pedro Domingos All rights reserved.
 * 
 * Contact: Jesse Davis, University of Washington
 * (jdavis@cs.washington.edu).
 * 
 * Redistribution and use in source and binary forms, with
 * or without modification, are permitted provided that
 * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above
 * copyright notice, this list of conditions and the
 * following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the
 * above copyright notice, this list of conditions and the
 * following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use
 * of this software must display the following
 * acknowledgment: "This product includes software
 * developed by Jesse Davis and Pedro in the 
 * Department of Computer Science and Engineering 
 * at the University of Washington".
 * 
 * 4. Your publications acknowledge the use or
 * contribution made by the Software to your research
 * using the following citation(s): 
 * Jesse Davis and Pedro Domingos, 
 * "Bottom-Up Learning of Markov Network Structure",
 * 27th International Conference on Machine Learning, 2010
 *
 * 
 * 5. Neither the name of the University of Washington nor
 * the names of its contributors may be used to endorse or
 * promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON
 * AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY
 * OF WASHINGTON OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

#ifndef FMRFLEARN_H_NOV_23_2005
#define FMRFLEARN_H_NOV_23_2005

#include <sys/time.h>
#include "rule.h"
#include "fpseudologlikelihood.h"
#include "flbfgsb.h"
#include <cmath>
#include <utility>

class FastMRFLearn{

 public:
 FastMRFLearn(Array<Array<int>*>* d, double pMean, double pDev, int kSize, double lPrior, int updateFrequency, int distanceFunc, bool pruneMLN, const string& outMLNFileName) :
  clusterSize(kSize), data(d), updateFreq(updateFrequency), priorMean(pMean), priorStdDev(pDev), litPrior(lPrior), distanceFunction(distanceFunc), prune(pruneMLN), outMLNFileName_(outMLNFileName)
  {
    id = 0;
    debug = false;
    trim = false;//true
    outputLevel = 0;

  }

  ~FastMRFLearn(){
    data->deleteItemsAndClear();
    //Array<Array<Array<Array<double>*>*>*> vdProb;
    vDiff.deleteItemsAndClear();
    delete data;
    delete pll;
    delete lbfgsb;
  }
  /*
  void run(bool unit, char* outFile, bool partialPrint){
    run(unit,outFile,partialPrint,false);
  }
  */
  void run(bool unit, char* outFile, bool partialPrint, bool fullData, int maxIter){
    buildValueDiff();
    Array<double>* exWt = new Array<double>;
    Array<Array<int>*>* sData = new Array<Array<int>*>;
    Array<Array<int>*>* coverList = new Array<Array<int>*>;
    Array<Rule*>* ruleSet = createRuleSet(data, sData,exWt, coverList, unit);
    kSizeArray.growToSize(10);
    kSizeArray[0] = 1;
    
    kSizeArray[1] = 2;
    kSizeArray[2] = 5;
    kSizeArray[3] = 10;
    kSizeArray[4] = 25;
    kSizeArray[5] = 50;
    kSizeArray[6] = 75;
    kSizeArray[7] = 100;
    kSizeArray[8] = 200;
    
    kSizeArray[9] = -1;//sData->size();
    pll = new FastPseudoLogLikelihood;
    lbfgsb =new FastLBFGSB(-1, -1, pll, ruleSet->size()); 
    lbfgsb->setMaxIter(maxIter); 
    lbfgsb->setFtol(0.01);
    pll->setExWt(exWt,coverList);
 
    //cout << "Here Run " << data->size() << " " << ruleSet->size() <<"\n";
    rise(ruleSet, data, sData,unit, fullData);

    //ruleSet->deleteItemsAndClear();
    for(int ix = 0; ix < ruleSet->size(); ix++){
      delete (*ruleSet)[ix];
      (*ruleSet)[ix] = NULL;
    }
    delete ruleSet;
    delete sData;
    coverList->deleteItemsAndClear();
    delete coverList;
    delete exWt;
  }



  Array<Rule*>* createRuleSet(Array<Array<int>*>* data, Array<Array<int>*>* smallData, Array<double>* exWt, Array<Array<int>*>* clist, bool unit){
    Array<Rule*>* ruleSet =new Array<Rule*>;
    for(int ix = 0; ix < data->size(); ix++){
      Rule* newRule = new Rule(-1);//idx, example);
      //Array<int>* example = new Array<int>;
      //IntHashArray* idx = new IntHashArray;
      Array<int>* cur = (*data)[ix];
      for(int jx = 0; jx < cur->size(); jx++){
	//example->append((*cur)[jx]);
	//idx->append(jx);
	if ((*cur)[jx] == 1){
	  newRule->addValue((*cur)[jx],jx);
	}
      }
      int idx = containsRule(newRule, ruleSet);
      if (idx < 0){
	newRule->setRid(ruleSet->size());
	ruleSet->append(newRule);
	smallData->append((*data)[ix]);
	exWt->append(1);
	Array<int>* tmp =new Array<int>;
	tmp->append(ix);
	clist->append(tmp);
      }
      else{
	(*exWt)[idx]++;
	Array<int>* tmp = (*clist)[idx];
	tmp->append(ix);
	delete newRule;
      }
    }
    if (unit){
      cout << "Adding Unit Clauses\n";
      Array<Rule*>* newRuleSet = new Array<Rule*>;
      for(int ix = 0; ix < (*data)[0]->size(); ix++){
	Rule* nRule = new Rule(-1);
	nRule->addValue(1, ix);
	newRuleSet->append(nRule);
      }
      for(int ix = 0; ix < ruleSet->size(); ix++){
	int nidx = containsRule((*ruleSet)[ix], newRuleSet);
	if (nidx < 0){
	  newRuleSet->append((*ruleSet)[ix]);
	}
	else{
	  delete (*ruleSet)[ix];
	}
      }
      delete ruleSet;
      return newRuleSet;
    }
    else{
      cout << "Not Adding Unit Clauses\n";
      return ruleSet;
    }
  }

  IntHashArray* newlyCovered(Rule* rule, Array<Array<int>*>* data){
    IntHashArray* covered = rule->covered;
    IntHashArray* nc = new IntHashArray;
    for(int ix = 0; ix < data->size(); ix++){
      if (!covered->contains(ix)){
	if (covers(rule, (*data)[ix])){
	  nc->append(ix);
      }
      }
    }
    return nc;
  }
  bool covers(Rule* rule, Array<int>* example){
    IntHashArray* ruleIdx = rule->index;
    Array<int>* ruleVal = rule->value;
    for(int ix = 0; ix < ruleIdx->size(); ix++){
      int pos = (*ruleIdx)[ix];
      
      if ((*ruleVal)[ix] != (*example)[pos]){
	return false;
      }
      
    }
    return true;
  }
  double hammingDistance(Rule* rule, Array<int>* example){
    double dist = 0;
    IntHashArray* ruleIdx = rule->index;
    Array<int>* ruleVal = rule->value;
    for(int ix = 0; ix < ruleIdx->size(); ix++){
      int pos = (*ruleIdx)[ix];
      
      if ((*ruleVal)[ix] != (*example)[pos]){
	dist++;
      }
      
    }
    return dist;
  }

  Rule* pickUnit(Rule* rule, IntHashArray* removedCond){
    IntHashArray* idx = new IntHashArray;
    IntHashArray* ruleIdx = rule->index;
    Array<int>* ruleVal = rule->value;
    Array<int>* newVal = new Array<int>;
    for(int ix = 0; ix < ruleIdx->size(); ix++){
      int pos = (*ruleIdx)[ix];
      
      if (ix == 0){
	idx->append(pos);
	newVal->append((*ruleVal)[ix]);
      }
      else{
	removedCond->append(pos);
      }
    }
    Rule* newRule = new Rule(idx,newVal);
    return newRule;
  }

  Rule* generalizeRule(Rule* rule, Array<int>* example, IntHashArray* removedCond){
    IntHashArray* idx = new IntHashArray;
    IntHashArray* ruleIdx = rule->index;
    Array<int>* ruleVal = rule->value;
    Array<int>* newVal = new Array<int>;
    
    for(int ix = 0; ix < ruleIdx->size(); ix++){
      int pos = (*ruleIdx)[ix];
      
      if ((*ruleVal)[ix] == (*example)[pos]){
	idx->append(pos);
	newVal->append((*example)[pos]);
      }
      else{
	removedCond->append(pos);
      }
      
    }
    Rule* newRule = new Rule(idx,newVal);
    newRule->setRid(rule->getRid());
    return newRule;
  }
  

  Rule* generalizeRule(Rule* rule, Array<Array<int>*>* examples, IntHashArray* removedCond){
    IntHashArray* idx = new IntHashArray;
    IntHashArray* ruleIdx = rule->index;
    Array<int>* ruleVal = rule->value;
    Array<int>* newVal = new Array<int>;
    for(int ix = 0; ix < ruleIdx->size(); ix++){
      int pos = (*ruleIdx)[ix];
      bool removed = false;
      for(int jx = 0; jx < examples->size() && !removed; jx++){
	Array<int>* example = (*examples)[jx];

	if ((*ruleVal)[ix] != (*example)[pos]){
	  removedCond->append(pos);
	  removed = true;
	}
      }
      if (!removed){
	idx->append(pos);
	newVal->append((*ruleVal)[ix]);
      }
    }
    Rule* newRule = new Rule(idx,newVal);
    newRule->setRid(rule->getRid());
    return newRule;
  }

  
  
  void find(Rule* rule, int numFeats, Array<Array<int>*>* data, int k,
	    Array<int>* index, bool fullData){
    //Array<int>* index = new Array<int>;
    Array<double> scores;
    Array<pair<double,int>*> fullScores;
    //index->growToSize(k);
    double minScore = numFeats; 
    int minIdx = -1;
    //cerr << "Enter\n";
    //cout << "Covered: ";
    /*
      cout << "Eval: ";
      rule->print(cout);
      cout << endl;
    */
    for(int ix = 0; ix < data->size(); ix++){
      Array<int>* example = (*data)[ix];
      bool covered = false;
      if (fullData){
	covered = rule->covered->contains(ix);
      }
      else{
	/*
	  for fastCov hack ??
	  if (ix != rule->getRid()){
	 */
	covered = rule->fastCov->contains(ix);
      }
      if (!covered){
	
	double newScore = 0;
	if (distanceFunction == 0){
	  newScore = hammingDistance(rule,example);
	}
	else if (distanceFunction == 1){
	  newScore = svdLimit(rule,example);//
	}
	else if (distanceFunction == 2){
	  newScore = svdLimitFeat(rule,example);//
	}
	else if (distanceFunction == 3){
	  newScore = svdLimitClass(rule,example);//
	}
	else if (distanceFunction == 4){
	  newScore = svd(rule,example);//
	}
	else{
	  
	}

	if (minScore > newScore || index->size() < k){
	  if (index->size() < k){
	    index->append(ix);
	    scores.append(newScore);
	    pair<double,int>* jpair = new pair<double,int>;
	    jpair->first = newScore;
	    jpair->second = ix;
	    fullScores.append(jpair);//new pair(newScore, ix));
	    if ( minScore < newScore || minIdx == -1){
	      minScore = newScore;
	      minIdx = index->size()-1;
	    }
	  }
	  else{
	    if (minIdx == -1 || minIdx >= index->size() 
		|| minIdx >= scores.size()){
	      cout << "HERE MIN BAD " << ix << " sz ";
	      cout << index->size() << " ";
	      cout << scores.size() << endl;
	      exit(-1);
	    }
	    fullScores[minIdx]->first = newScore;
	    fullScores[minIdx]->second = ix;
	    scores[minIdx] = newScore;
	    (*index)[minIdx] = ix;
	    minIdx = 0;
	    minScore = scores[0];
	    for(int jx = 1; jx < index->size(); jx++){
	      if (scores[jx] > minScore){
		minScore = scores[jx];
		minIdx = jx;
	      }
	    }
	  }
	}
      }
 
    }
 
    //pair<int,int>* jpair = new pair<int,int>;
    pair<double,int>* tempItem;
    for (int i = 0; i < fullScores.size(); i++)
      for (int j = i+1; j < fullScores.size(); j++)
        if ((fullScores[i]->first > fullScores[j]->first) ||
	    (fullScores[i]->first == fullScores[j]->first &&
	     fullScores[i]->second > fullScores[j]->second))
	  {
          tempItem = fullScores[i];
          fullScores[i] = fullScores[j];
          fullScores[j] = tempItem;
        }

    //fullScores.quicksort();
      //fullScores.growToSize(scores.size();
    for(int ix = 0; ix < fullScores.size(); ix++){
      (*index)[ix] = fullScores[ix]->second;
     
    }
    fullScores.deleteItemsAndClear();
  }

  
  void rise(Array<Rule*>* ruleSet, Array<Array<int>*>* data, Array<Array<int>*>* sData, bool unit, bool fullData){
    
    //printRuleSet(ruleSet); 
    Timer timer;
    double iTime = timer.time();
    Array<int*>* clauseIdx = new Array<int*>;
    clauseIdx->growToSize(ruleSet->size());
    
    for(int ix = 0; ix < clauseIdx->size(); ix++){
      int* tmpIdx = new int;
      *tmpIdx = ix;
      (*clauseIdx)[ix] = tmpIdx;
    }
    long double structurePrior = 0;
    for(int ix = 0; ix < ruleSet->size(); ix++){
      structurePrior += (*ruleSet)[ix]->index->size() * litPrior;
    }
    cout << "Enter Fast (w/compression) RISE\n";
    Array<double> priorMeans, priorStdDevs;
    bool improving = true;
    pllSetPriorMeansStdDevs(ruleSet->size(), priorMeans, priorStdDevs);

    int numFeats = (*data)[0]->size();
    //pll->setInclude(numFeats,sData->size());
    cout << "Begin init " << numFeats << "\n";
    pll->initializeCounts(sData,ruleSet, clauseIdx);
    
    bool error;
    int iter;
    cout << "Initialized ";
    timer.printTime(cout, timer.time()-iTime); cout << endl;
    Array<double>* weights = new Array<double>;
  
    Array<double>* tmpWeights = new Array<double>;
    weights->growToSize(ruleSet->size()+1, 0);
    tmpWeights->growToSize(ruleSet->size()+1, 0);
    for(int ix = 0; ix < weights->size()-1; ix++){
      (*weights)[ix+1] = 0;
    }
    
    double* wwts = (double*) weights->getItems();
    //pll->setJUnify(true);
    pll->setJ(true);
    long double iterScore = lbfgsb->minimize(ruleSet->size(), wwts, iter, error);
    cout << "Init Score = " << (iterScore+structurePrior) << " ";
    cout << iterScore << " prior " << structurePrior << " Rule Set Size ";
    cout << ruleSet->size() << endl; //" Iter " << iter << endl;
    cout << "K Size " << clusterSize << " Lit Prior " << litPrior << " Update Frequency " << updateFreq << endl;
    if (distanceFunction == 0){
      cout << "Distance Function: Hamming\n";
    }
    else if (distanceFunction == 1){
      cout << "Distance Function: Only Attributes/Class from Rule\n";
    }
    else if (distanceFunction == 2){
      cout << "Distance Function: Only Attributes from Rule (all class)\n";
    }
    else if (distanceFunction == 3){
      cout << "Distance Function: Only class from Rule (all attributes)\n";
    }
    else if (distanceFunction == 4){
      cout << "Distance Function: Full Attributes & Class\n";
    }
    
    iterScore += structurePrior;
    long double currScore = iterScore;
    for(int ix = 0; ix < weights->size(); ix++){
      (*tmpWeights)[ix] = (*weights)[ix];
    }
    Array<Array<int>*>* points = new Array<Array<int>*>;
    
    
    double startTime = timer.time();
    if (debug)
      printRuleSetAndWts(ruleSet, weights);
    //exit(-1);
    int fullIter = 1;
    IntHashArray* removedCond = new IntHashArray;
    Array<int>* idx = new Array<int>;

    pll->setFastPLL(true);
    int totalRemoves = 0;
    int totalUpdates = 0;
    int rulesScored = 0;
    Array<Rule*>* builtRules = new Array<Rule*>;
    bool seen24hr = false;
    bool seen48hr = false;
    /*
    for(int ix = 0; ix < ruleSet->size(); ix++){
      cout << "ix " << ix << ": " << (*weights)[ix+1] << " "; 
      (*ruleSet)[ix]->printBrief(cout);
      cout << " idx " << *(*clauseIdx)[ix] << endl;
    }
    */
    while(improving){
      double begSec_ = timer.time();
      cout << "Top Level Iteration: " << fullIter << endl;
      int start = 0;
      if (unit){
	start = numFeats;
      }
      improving = false;
      for(int ix = start; ix < ruleSet->size(); ){
	
	cout << "Rule iter: " << ix << " of " << ruleSet->size() << endl;
	/*
	  cout << "Init: ";
	  (*ruleSet)[ix]->printBrief(cout);
	  cout << " idx " << *(*clauseIdx)[ix] << endl;
	  cout << "Weight: " << (*weights)[9] << " " << (*tmpWeights)[9] <<endl;
	*/
	int bestKSize = -1;
	long double bestKScore = -1;
	idx->clear();
	double begFtime = timer.time();
	if (!seen24hr){
	  //86400
	  if ((begFtime - startTime) > 86400){ 
	    seen24hr = true;
	    printMLNToFile("24hr",-2,(*sData)[0]->size(),ruleSet,weights);
	  }
	}
	else if (!seen48hr){
	  if ((begFtime - startTime) > 172800){ 
	    seen48hr = true;
	    printMLNToFile("48hr",-2,(*sData)[0]->size(),ruleSet,weights);
	  }
	}

	if (fullData){
	  find((*ruleSet)[ix], numFeats, data, kSizeArray[kSizeArray.size()-2], idx, fullData);
	}
	else{
	  find((*ruleSet)[ix], numFeats, sData, kSizeArray[kSizeArray.size()-2], idx, fullData);
	}

	bool pickedRemove = false;
	builtRules->deleteItemsAndClear();
	bool scoredRemove = false;
	begFtime = timer.time();
	int gRuleSz = -1;

	for(int kx = 0; kx < kSizeArray.size(); kx++){
	  if (kx == (kSizeArray.size()-1)){

	  }
	  else{
	    //issue if mult rules have same distance, but different generalizatiosn
	    int kLoopBound = kSizeArray[kx];
	    if (kLoopBound > idx->size()){
	      kLoopBound = idx->size();
	    }
	    points->clear();
	    points->growToSize(kLoopBound);
	    for(int jx = 0; jx < kLoopBound; jx++){
	      if (fullData){
		(*points)[jx] = (*data)[(*idx)[jx]];
	      }
	      else{
		(*points)[jx] = (*sData)[(*idx)[jx]];
	      }
	    }
	  }
	  if ((idx->size() > 0) || (kx == (kSizeArray.size()-1))){

	    removedCond->clear();
	    Rule* genRule;
	    if (kx == (kSizeArray.size()-1)){
	      genRule = pickUnit((*ruleSet)[ix], removedCond);
	    }
	    else{
	      genRule = generalizeRule((*ruleSet)[ix], points, removedCond);
	    }
	    bool alreadyScored = (containsRule(genRule,builtRules) > -1);
	    	    
	    if (debug){
	      (*ruleSet)[ix]->print(cout);
	      cout << " -> ";
	      genRule->print(cout);
	      cout << endl;
	    }
	    long double newPrior = structurePrior;
	    int dupRuleIdx = containsRule(genRule, ruleSet);
	    bool removeRule = (dupRuleIdx > -1);
	    if (!alreadyScored && !scoredRemove){

	      rulesScored++;
	      scoredRemove = removeRule;
	      (*tmpWeights)[dupRuleIdx+1] = 0;
	      double begTmp = timer.time();
	      if (removeRule){
		pll->setIgnoreSet(removedCond, (*ruleSet)[ix], (*ruleSet)[dupRuleIdx], dupRuleIdx, false);
		
		newPrior = newPrior - ((*ruleSet)[ix]->index->size() * litPrior);
		if (outputLevel > 1)
		  cout << "NewPrior " << newPrior << " = " << structurePrior << " - " << (*ruleSet)[ix]->index->size() << endl;
	      }
	      else{
		(*tmpWeights)[ix+1] = 0;
		pll->countNewRule(genRule, (*ruleSet)[ix], (*clauseIdx)[ix], sData, removedCond, false);
		newPrior = structurePrior - 
		  ((*ruleSet)[ix]->index->size()*litPrior) +
		  (genRule->index->size() * litPrior);
		if (outputLevel > 1){
		  cout << "NewPrior " << newPrior << " = ";
		  cout << structurePrior << " - " << (litPrior * (*ruleSet)[ix]->index->size());
		  cout << " + " << (litPrior * genRule->index->size()) << ": " << 
		    litPrior << " " << (*ruleSet)[ix]->index->size() << " " << genRule->index->size()  << endl;
		  cout << "NewPrior " << newPrior << " = " << structurePrior;
		  cout << " - " << (*ruleSet)[ix]->index->size();
		  cout << " + " << genRule->index->size() << endl;
		}
	      }
	      /*
	      cout << "K " << kSizeArray[kx] << " ";
	      genRule->printBrief(cout);
	      cout << " idx " << *(*clauseIdx)[ix] << endl;
	      */
	      if (outputLevel > 0){
		cout << "Counts: ";
		timer.printTime(cout, timer.time()-begTmp); cout << endl;
	      }
	      //learn new weights	  

	      double* nwts = (double*) tmpWeights->getItems();
	      begTmp = timer.time();
	      //pll->setJ(true);
	      long double tmpScore = lbfgsb->minimize(ruleSet->size(), nwts, iter, error, ix);
	      if (false){
		if (genRule->index->size() > 0){
		  int checkIdx = (*genRule->index)[0];
		  cout << "New score: " << (tmpScore+newPrior) << " = " << tmpScore << " + " << newPrior << " Remove " << removeRule << " Iter " << fullIter << " " << ix << " " << currScore << endl;
		  cerr << "check: " << checkIdx << " ";
		  cout << (*weights)[checkIdx+1] << " " << (*tmpWeights)[checkIdx+1] << " : ";
		}
		cout << (*weights)[ix] << " " << (*tmpWeights)[ix] << " : ";
		cout << (*weights)[ix+1] << " " << (*tmpWeights)[ix+1] << endl;
	      }

	      for(int rx = 0; rx < weights->size(); rx++){
		(*tmpWeights)[rx] = (*weights)[rx];
	      }
	      //cout << "New score: " << (tmpScore+newPrior) << " = " << tmpScore << " + " << newPrior << " Remove " << removeRule << " Iter " << fullIter << " " << ix << " " << currScore << endl;
	     
	      tmpScore += newPrior;

	      if (bestKSize == -1){
		bestKSize = kx;
		bestKScore = tmpScore;
		pickedRemove = removeRule;
		gRuleSz = genRule->index->size();
	      }
	      else if (tmpScore < bestKScore){
		bestKSize = kx;
		bestKScore = tmpScore;
		pickedRemove = removeRule;
		gRuleSz = genRule->index->size();
	      }
	      builtRules->append(genRule);
	      //delete genRule;
	      if (outputLevel > 0){
		cout << "LBFGS took: ";
		timer.printTime(cout, timer.time()-begTmp); 
		cout << " Iter: " << iter << endl;// << endl;
	      }
	    }//builtRules contains
	    else{
	      /*
	      if (alreadyScored){
		cout << "KX " << kx << ": Already Scored!!\n";
	      }
	      else{
		cout << "KX " << kx << ": Already Scored (Remove)!!\n";
	      }
	      */
	      delete genRule;
	    }
	  }//idx
	}//end loop through kSizes
	//cerr << "Delete builtRules " << endl;
	builtRules->deleteItemsAndClear();
	//cerr << "Finished Delete builtRules " << endl;
	//cout << "Total eval time ";
	//timer.printTime(cout, timer.time()-begFtime); cout << endl;
	//begFtime = timer.time();
	if ((bestKSize > -1) && (bestKScore < currScore)){
	  
	  points->clear();
	  if (bestKSize == (kSizeArray.size()-1)){

	  }
	  else{
	    points->growToSize(kSizeArray[bestKSize]);
	    for(int jx = 0; jx < kSizeArray[bestKSize]; jx++){
	      if (fullData){
		(*points)[jx] = (*data)[(*idx)[jx]];
	      }
	      else{
		(*points)[jx] = (*sData)[(*idx)[jx]];
	      }
	    }
	  }
	  if (idx->size() > 0){
	    removedCond->clear();
	    Rule* genRule;
	    if (bestKSize == (kSizeArray.size()-1)){
	      genRule = pickUnit((*ruleSet)[ix], removedCond);
	    }
	    else{
	      genRule = generalizeRule((*ruleSet)[ix], points, removedCond);
	    }
	    //Rule* genRule = generalizeRule((*ruleSet)[ix], points, removedCond);
	    /*
	      cout << "Best K " << bestKSize << " ";
	      genRule->printBrief(cout);
	      cout << " idx " << *(*clauseIdx)[ix] << endl;
	    */
	    if (debug){
	      (*ruleSet)[ix]->print(cout);
	      cout << " -> ";
	      genRule->print(cout);
	      cout << endl;
	    }
	    long double newPrior = structurePrior;
	    int dupRuleIdx = containsRule(genRule, ruleSet);
	    bool removeRule = (dupRuleIdx > -1);
	    
	    double begTmp = timer.time();
	    double* nwts = (double*) tmpWeights->getItems();
	    if (removeRule){
	      (*tmpWeights)[dupRuleIdx+1] = 0;
	      pll->setIgnoreSet(removedCond, (*ruleSet)[ix], (*ruleSet)[dupRuleIdx], dupRuleIdx, true);
	      newPrior = newPrior - ((*ruleSet)[ix]->index->size() * litPrior);
	      if (outputLevel > 1)
		cout << "NewPrior " << newPrior << " = " << structurePrior << " - " << (*ruleSet)[ix]->index->size() << endl;
	    }
	    else{
	      (*tmpWeights)[ix+1] = 0;
	      pll->countNewRule(genRule, (*ruleSet)[ix], (*clauseIdx)[ix], sData, removedCond, true);
	      newPrior = structurePrior - 
		((*ruleSet)[ix]->index->size()*litPrior) +
		(genRule->index->size() * litPrior);
	      if (outputLevel > 1){
		cout << "NewPrior " << newPrior << " = ";
		cout << structurePrior << " - " << (litPrior * (*ruleSet)[ix]->index->size());
		cout << " + " << (litPrior * genRule->index->size()) << ": " << 
		  litPrior << " " << (*ruleSet)[ix]->index->size() << " " << genRule->index->size()  << endl;
		cout << "NewPrior " << newPrior << " = " << structurePrior;
		cout << " - " << (*ruleSet)[ix]->index->size();
		cout << " + " << genRule->index->size() << endl;
	      }
	    }
	    
	    if (outputLevel > 0){
	      cout << "Counts: ";
	      timer.printTime(cout, timer.time()-begTmp); cout << endl;
	    }
	    //learn new weights	  

	    
	    begTmp = timer.time();
	    //double tmpScore = lbfgsb->minimize(ruleSet->size(), nwts, iter, error);//, ix);
	    //pll->setJUnify(true);
	    pll->setJ(true);
	    long double tmpScore = lbfgsb->minimize(ruleSet->size(), nwts, iter, error, ix);
	    
	    cout << "Picked: " << kSizeArray[bestKSize] << " New Score " << (tmpScore+newPrior) << " " << tmpScore << " prior " << newPrior << " ";
	    cout << " f sz " << (*ruleSet)[ix]->index->size() << " -> ";
	    cout << genRule->index->size() << " ";
	    if (pickedRemove){
	      cout << "Remove = true" << endl ;//<< " iter ";
	    }
	    else{
	      cout << "Remove = false" << endl;
	    }
	    tmpScore += newPrior;
	  
	    if (outputLevel > 0){
	      cout << "LBFGS took: ";
	      timer.printTime(cout, timer.time()-begTmp); 
	      cout << " Iter: " << iter << endl;// << endl;
	    }
	    if (tmpScore < currScore){
	      totalUpdates++;
	      improving = true;
	      structurePrior = newPrior;
	      /*
		cout << "Before update model: ";
		pll->print(695);
	      */
	      pll->updateModel(removeRule, ix, genRule->getRid());
	      /*
		cout << "After update model: ";
		pll->print(695);
	      */
	      cout << "Update model " << ix << ": New Score " << tmpScore << " Iteration: " << fullIter << " with K Size " << kSizeArray[bestKSize] << "\n";
	      if (removeRule){
		totalRemoves++;
		cout << "Remove Rule " << ix << " Rule Set size = " << ruleSet->size() << endl;
		Rule* discard = ruleSet->removeItemFastDisorder(ix);
		int * ruleIdx = clauseIdx->removeItemFastDisorder(ix);
		//cerr << "Discard RuleIdx " << *((*clauseIdx)[ix]) << " Old " << *ruleIdx << endl;
		delete ruleIdx;
		delete genRule;
		//flag ???
		if (ix < ruleSet->size()){
		  *(*clauseIdx)[ix] = ix;
		}
		delete discard;
		/*
		  cout << "After discard: ";
		  pll->print(695);
		*/
		if ((totalUpdates % updateFreq) == 0){
		  (*weights)[dupRuleIdx+1] = 0;
		}
		
		tmpWeights->removeItemFastDisorder(ix+1);
		weights->removeItemFastDisorder(ix+1);
		pll->removeUpdate(ix);
		/*
		  cout << "After remove update: ";
		  pll->print(695);
		*/
	      }
	      else{
		Rule* discard = (*ruleSet)[ix];
		if ((totalUpdates % updateFreq) == 0){
		  (*weights)[ix+1] = 0;
		}
		(*ruleSet)[ix] = genRule;
		delete discard;
	      }
	      if ((totalUpdates % updateFreq) == 0){
		nwts = (double*) weights->getItems();
		//pll->setJUnify(true);
		pll->setJ(true);
		long double fullScore = lbfgsb->minimize(ruleSet->size(), nwts, iter, error);
	  
		if (outputLevel > 0){
		  cout << "Full run of LBFGS took: ";
		  timer.printTime(cout, timer.time()-begTmp); 
		  cout << " Iter: " << iter << endl;// << endl;
		}
		currScore = fullScore + structurePrior;
		for(int jx = 0; jx < weights->size(); jx++){
		  (*tmpWeights)[jx] = (*weights)[jx];
		}
	      }
	      else{
		pll->updateCache();
		currScore = tmpScore;
		for(int jx = 0; jx < weights->size(); jx++){
		  (*weights)[jx] = (*tmpWeights)[jx];
		}
	      }
	      
	      if (!removeRule){
		ix++;
	      }
	      if (debug){
		printRuleSetAndWts(ruleSet, weights);
		cout << endl;
	      }
	    }
	    else{
	      for(int jx = 0; jx < weights->size(); jx++){
		(*tmpWeights)[jx] = (*weights)[jx];
	      }
	      ix++;
	      delete genRule;
	    }
	  }//end of ix
	  else{
	    ix++;
	  }
	}
	else{
	  if (bestKScore > currScore){
	    cout << "Picked: " << kSizeArray[bestKSize] << " New Score " << bestKScore << " " << currScore;
	    cout << " f sz " << (*ruleSet)[ix]->index->size() << " -> " << gRuleSz << endl;
	  }
	  ix++;
	}
	cout << endl;
	//cout << "Total update time ";
	//timer.printTime(cout, timer.time()-begFtime); cout << endl;
      }
      cout << "Iteration took: ";
      timer.printTime(cout, timer.time()-begSec_); 
      cout << endl;
      cout << "Iter Score " << iterScore << " " << currScore << endl << endl;
      if (currScore < iterScore){
	iterScore = currScore;
      }
      else{
	//improving = false;
      }
      fullIter++;
    }
    delete removedCond;
    cout << "Total time: ";
    timer.printTime(cout, timer.time()-startTime); 
    cout << endl;


    cout << "Final Rules: " << ruleSet->size() << " Total Updates " << totalUpdates << " Total Removes " << totalRemoves << " Total Rules Scored " << rulesScored << " In " << (fullIter-1) << " Iterations" << endl;
    //printRuleSetAndWts(ruleSet, weights);
    cout << "Train set score: " << iterScore << endl;
    printMLNToFile(NULL,-2,(*sData)[0]->size(),ruleSet,weights);

    delete builtRules;
    delete points;
    delete idx;   
    //points->deleteItemsAndClear();
    //delete points;
    //Array<int*>* clauseIdx = new Array<int*>;
    for(int ix = 0; ix < clauseIdx->size(); ix++){
      if ((*clauseIdx)[ix]){
	int * tmpInt = (*clauseIdx)[ix];
	delete tmpInt;
	(*clauseIdx)[ix] =NULL;
      }
    }
    //clauseIdx->deleteItemsAndClear();
    delete clauseIdx;
    delete weights;
    delete tmpWeights;

  }

  void pllSetPriorMeansStdDevs(int size, Array<double>& priorMeans, Array<double>& priorStdDevs)
    /*
      Array<double>& priorMeans, 
      Array<double>& priorStdDevs, const int& addSlots,
      const Array<int>* const & removedSlotIndexes)
    */
  {
    if (true)
    {
      /*
      setPriorMeansStdDevs(priorMeans, priorStdDevs, 
                           addSlots, removedSlotIndexes);
      */
      priorMeans.clear();
      priorStdDevs.clear();
      priorMeans.growToSize(size, priorMean);
      priorStdDevs.growToSize(size, priorStdDev);

      pll->setMeansStdDevs(priorMeans.size(), priorMeans.getItems(), 
                            priorStdDevs.getItems());
    }
    else
      pll->setMeansStdDevs(-1, NULL, NULL);
  }

  int containsRule(Rule* rule, Array<Rule*>* ruleSet){
    //int idx = -1;
    for(int ix = 0; ix < ruleSet->size(); ix++){
      if (rule->equals((*ruleSet)[ix])){
	return ix;
      }
    }
    return -1;
  }

  void printRuleSet(Array<Rule*>* ruleSet){
    for(int ix = 0; ix < ruleSet->size(); ix++){
      cout << ix << ": ";
      (*ruleSet)[ix]->print(cout);
      cout << endl;
    }
  }

  void printRuleSetAndWts(Array<Rule*>* ruleSet, Array<double>* wts){
    int ruleIdx = 0;
    for(int ix = 0; ix < ruleSet->size(); ix++){
      if ((*ruleSet)[ix]->index->size() > 0){
	cout << ruleIdx << ": ";
	cout << (*wts)[ix+1] << " : ";
	(*ruleSet)[ix]->print(cout);
	cout << endl;
	ruleIdx++;
      }
    }
  }

  ostream& printRuleSetAndWts(Array<Rule*>* ruleSet, Array<double>* wts, ostream& out) const
    {
      for(int ix = 0; ix < ruleSet->size(); ix++){
	if ((*ruleSet)[ix]->index->size() > 0){
	  if (prune && (*ruleSet)[ix]->covered->size() > 1){
	    out << (*wts)[ix+1] << " ";
	    (*ruleSet)[ix]->printMLN(out);
	    out << endl;
	  }
	  else{
	    out << (*wts)[ix+1] << " ";
	    (*ruleSet)[ix]->printMLN(out);
	    out << endl;
	  }
	}
      }
      return out;
    }
  



  void printMLNToFile(const char* const & appendStr, const int& iter,
		      int numPreds, Array<Rule*>* ruleSet, Array<double>* wts)
  {
    string fname = outMLNFileName_;

    if (appendStr) fname.append(".").append(appendStr);

    if (iter >= -1) 
    { 
      char buf[100]; 
      sprintf(buf, "%d", iter);
      fname.append(".iter").append(buf);
    }

    if (appendStr || iter >= -1) fname.append(".mln");
    
    ofstream out(fname.c_str());
    if (!out.good()) { cout << "ERROR: failed to open " <<fname<<endl;exit(-1);}

      // output the predicate declaration
    out << "//predicate declarations" << endl;
    for(int ix = 0; ix < numPreds; ix++){
      out << "P" << ix << "(id)\n";
    }
    out << endl;
    printRuleSetAndWts(ruleSet, wts, out);


    out << endl;
    out.close();
  } 

  double svd(Rule* rule, Array<int>* example){
    double score = 0;
    IntHashArray* cond = rule->index;
    //Array<int>* val = rule->value;
    ///Array<Array<Array<double>*>*>* v0 = vdProb[0];
    //A/rray<Array<Array<double>*>*>* v1 = vdProb[1];
    for(int ix = 0; ix < example->size(); ix++){
      //int pos = (*cond)[ix];
      if (((*example)[ix] == 1 && !cond->contains(ix)) ||
	  ((*example)[ix] == 0 && cond->contains(ix))){
	for(int cx = 0; cx < example->size(); cx++){
	  if (cx != ix){
	    //abs();
	    //hardwired binary
	    score += (*(vDiff[cx]))[ix];
	    /*
	      abs( ((*((*((*v0)[1]))[cx]))[ix]) - 
			  ((*((*((*v0)[0]))[cx]))[ix]));
	    
	    cout << ((*((*((*v0)[1]))[cx]))[ix]) << "-" <<
	      ((*((*((*v0)[0]))[cx]))[ix]) << " " << cx << " " << ix << endl;
	    cout << ((*((*((*v1)[1]))[cx]))[ix]) << "-" <<
	      ((*((*((*v1)[0]))[cx]))[ix]) << " " << cx << " " << ix << endl << endl;
	    
	      score += abs( ((*((*((*v1)[1]))[cx]))[ix]) - 
	      ((*((*((*v1)[0]))[cx]))[ix]));
	    */
	  }
	}
      } 
    }
    return score;
  }
  double svdLimitFeat(Rule* rule, Array<int>* example){
    double score = 0;
    IntHashArray* cond = rule->index;
    //Array<int>* val = rule->value;
    //Array<Array<Array<double>*>*>* v0 = vdProb[0];
    //Array<Array<Array<double>*>*>* v1 = vdProb[1];
    for(int ix = 0; ix < cond->size(); ix++){
      int pos = (*cond)[ix];
      if (((*example)[pos] == 0)){
	for(int cx = 0; cx < example->size(); cx++){
	  if (cx != pos){
	    //abs();
	    //hardwired binary
	    score += (*(vDiff[cx]))[pos];
	    /*
	      abs( ((*((*((*v0)[1]))[cx]))[pos]) - 
	      ((*((*((*v0)[0]))[cx]))[pos]));
	      
	    cout << ((*((*((*v0)[1]))[cx]))[ix]) << "-" <<
	      ((*((*((*v0)[0]))[cx]))[ix]) << " " << cx << " " << ix << endl;
	    cout << ((*((*((*v1)[1]))[cx]))[ix]) << "-" <<
	      ((*((*((*v1)[0]))[cx]))[ix]) << " " << cx << " " << ix << endl << endl;
	    
	    score += abs( ((*((*((*v1)[1]))[cx]))[pos]) - 
			  ((*((*((*v1)[0]))[cx]))[pos]));
	    */
	  }
	}
      } 
    }
    return score;
  }

  double svdLimit(Rule* rule, Array<int>* example){
    double score = 0;
    IntHashArray* cond = rule->index;
    //Array<int>* val = rule->value;
    //Array<Array<Array<double>*>*>* v0 = vdProb[0];
    //Array<Array<Array<double>*>*>* v1 = vdProb[1];
    for(int ix = 0; ix < cond->size(); ix++){
      int pos = (*cond)[ix];
      if (((*example)[pos] == 0)){
	for(int cx = 0; cx < cond->size(); cx++){
	  int cPos = (*cond)[cx];
	  if (cPos != pos){
	    //abs();
	    //hardwired binary
	    score += (*(vDiff[cPos]))[pos];
	    //abs( ((*((*((*v0)[1]))[cPos]))[pos]) - 
	    //((*((*((*v0)[0]))[cPos]))[pos]));
	    /*
	    cout << ((*((*((*v0)[1]))[cx]))[ix]) << "-" <<
	      ((*((*((*v0)[0]))[cx]))[ix]) << " " << cx << " " << ix << endl;
	    cout << ((*((*((*v1)[1]))[cx]))[ix]) << "-" <<
	      ((*((*((*v1)[0]))[cx]))[ix]) << " " << cx << " " << ix << endl << endl;
	    
	    score += abs( ((*((*((*v1)[1]))[cPos]))[pos]) - 
			  ((*((*((*v1)[0]))[cPos]))[pos]));
	    */
	  }
	}
      } 
    }
    return score;
  }
  double svdLimitClass(Rule* rule, Array<int>* example){
    double score = 0;
    IntHashArray* cond = rule->index;

    for(int ix = 0; ix < example->size(); ix++){
      //int pos = (*cond)[ix];
      if (!cond->contains(ix)){
	for(int cx = 0; cx < cond->size(); cx++){
	  int cPos = (*cond)[cx];
	  if (cPos != ix){
	    //hardwired binary
	    score += (*(vDiff[cPos]))[ix];

	  }
	}
      } 
    }
    return score;
  }

  void buildValueDiff(){
    Array<Array<Array<Array<double>*>*>*> vdProb;
    int numFeats = (*data)[0]->size();
    vdProb.growToSize(2);
    vdProb[0] = new Array<Array<Array<double>*>*>;
    vdProb[0]->growToSize(2);
    
    vdProb[1] = new Array<Array<Array<double>*>*>;
    vdProb[1]->growToSize(2);
    (*(vdProb[0]))[0] = new Array<Array<double>*>;
    (*(vdProb[0]))[1] = new Array<Array<double>*>;
    (*(vdProb[1]))[0] = new Array<Array<double>*>;
    (*(vdProb[1]))[1] = new Array<Array<double>*>;
    for(int ix = 0; ix < numFeats; ix++){
      Array<double>* nDiff =new Array<double>;
      nDiff->growToSize(numFeats,0);
      vDiff.append(nDiff);
    }
    Array<Array<double>*>* tmp = (*(vdProb[0]))[0];//new Array<Array<double>*>;
    Array<double>* t1;
    for(int jx = 0; jx < numFeats; jx++){
      t1 = new Array<double>;
      t1->growToSize(numFeats,1);
      tmp->append(t1);
    }
    //(*(vdProb[0]))[0]->append(tmp);
    
    tmp = (*(vdProb[0]))[1];//new Array<Array<double>*>;
    for(int jx = 0; jx < numFeats; jx++){
      t1 = new Array<double>;
      t1->growToSize(numFeats,1);	
      tmp->append(t1);
    }
    //(*(vdProb[0]))[1]->append(tmp);
    
    tmp =(*(vdProb[1]))[0];
    for(int jx = 0; jx < numFeats; jx++){
      t1 = new Array<double>;
      t1->growToSize(numFeats,1);	
      tmp->append(t1);
    }
    
    
    tmp = (*(vdProb[1]))[1];
    for(int jx = 0; jx < numFeats; jx++){
      t1 = new Array<double>;
      t1->growToSize(numFeats,1);
      tmp->append(t1);	
    }
    
    for(int cx = 0; cx < numFeats; cx++){
      for(int ix = 0; ix < data->size(); ix++){
	Array<int>* example = (*data)[ix];
	int classVal = (*example)[cx];
	Array<Array<Array<double>*>*>* v1 = vdProb[classVal];
	for(int jx = 0; jx < example->size(); jx++){
	  if (jx != cx){
	    if ((*example)[jx] == 1){
	      (*((*((*v1)[1]))[cx]))[jx]++;// = NULL;
	    }
	    else{
	      (*((*((*v1)[0]))[cx]))[jx]++;// = NULL;
	    }
	  }
	}
      }
    }
    Array<Array<Array<double>*>*>* v0 = vdProb[0];
    Array<Array<Array<double>*>*>* v1 = vdProb[1];
    for(int ix = 0; ix < numFeats; ix++){
      for(int jx = 0; jx < numFeats; jx++){
	double norm = ((*((*((*v0)[0]))[ix]))[jx]) + ((*((*((*v1)[0]))[ix]))[jx]);
	
	//cout << ((*((*((*v0)[0]))[ix]))[jx]) << " " << ((*((*((*v1)[0]))[ix]))[jx]) << " -> ";
	((*((*((*v0)[0]))[ix]))[jx]) = ((*((*((*v0)[0]))[ix]))[jx]) / norm;
	((*((*((*v1)[0]))[ix]))[jx]) = ((*((*((*v1)[0]))[ix]))[jx]) / norm;
	//cout << ((*((*((*v0)[0]))[ix]))[jx]) << " " << ((*((*((*v1)[0]))[ix]))[jx]) << " " << ix << " " << jx << endl;
	//cout << ((*((*((*v0)[0]))[ix]))[jx]) << " p(v" << ix << " = 0 | " << "v" << jx << " = 0) " << ((*((*((*v1)[0]))[ix]))[jx]) << " p(v" << ix << " = 1 | v" << jx << " = 0)" << endl;
	norm = ((*((*((*v0)[1]))[ix]))[jx]) + ((*((*((*v1)[1]))[ix]))[jx]);
	//cout << ((*((*((*v0)[1]))[ix]))[jx]) << " " << ((*((*((*v1)[1]))[ix]))[jx]) << " -> ";
	((*((*((*v0)[1]))[ix]))[jx]) = ((*((*((*v0)[1]))[ix]))[jx]) / norm;
	((*((*((*v1)[1]))[ix]))[jx]) = ((*((*((*v1)[1]))[ix]))[jx]) / norm;
	double s1 = abs( ((*((*((*v1)[1]))[ix]))[jx]) - 
			 ((*((*((*v1)[0]))[ix]))[jx]));
	double s2 = abs( ((*((*((*v0)[1]))[ix]))[jx]) - 
			 ((*((*((*v0)[0]))[ix]))[jx]));
	(*(vDiff[ix]))[jx] = s1 + s2;
	//cout << ((*((*((*v0)[1]))[ix]))[jx]) << " p(v" << ix << " = 0 | " << "v" << jx << " = 1) " << ((*((*((*v1)[1]))[ix]))[jx]) << " p(v" << ix << " = 1 | v" << jx << " = 1)" << endl << endl;	
      }
    }
    //cout << vDiff.size() << " " << vDiff[0]->size() << endl;
    //cout << "v0 " << v0->size() << endl;
    for(int ix = 0; ix < v0->size(); ix++){
      
      Array<Array<double>*>* tmp1 = (*v0)[ix];
      //cout << "tmp1 " << tmp1->size() << endl;
      for(int jx = 0; jx < tmp1->size(); jx++){
	//(*tmp1)[jx]->deleteItemsAndClear();
	delete (*tmp1)[jx];
	(*tmp1)[jx] = NULL;
      }
      delete tmp1;
      (*v0)[ix] = NULL;
    }
    delete v0;
    for(int ix = 0; ix < v1->size(); ix++){
      Array<Array<double>*>* tmp1 = (*v1)[ix];
      tmp1->deleteItemsAndClear();
      delete tmp1;
      (*v1)[ix] = NULL;
    }
    delete v1;

  }

 private:
  //Array<int>* clusterSize;
  Array<int>  kSizeArray;
  int clusterSize;
  bool debug;
  bool trim;
  Array<Array<int>*>* data;
  Array<Array<double>*> vDiff;
  //Array<Array<double>*> negClass;
  
  FastPseudoLogLikelihood* pll;
  FastLBFGSB* lbfgsb;
  int updateFreq;
  int outputLevel;
  double priorMean;
  double priorStdDev;
  double litPrior;
  int distanceFunction;
  bool prune;
  string outMLNFileName_;
  //bool fastPll;
  int id;
};

#endif
