/*
 * All of the documentation and software included in the
 * BLM Software is copyrighted by Jesse Davis and Pedro Domingos.
 * The arguments.cpp, arguments.h, hasharray.h, hash.h, hashint.h
 * flbfgsb.h, lbfgsb.h, random.h and timer.h come from the
 * Alchemy Software system. These files are denoted as coming from
 * Alchemy.
 * 
 * Copyright [2010] Jesse Davis and Pedro Domingos All rights reserved.
 * 
 * Contact: Jesse Davis, University of Washington
 * (jdavis@cs.washington.edu).
 * 
 * Redistribution and use in source and binary forms, with
 * or without modification, are permitted provided that
 * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above
 * copyright notice, this list of conditions and the
 * following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the
 * above copyright notice, this list of conditions and the
 * following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use
 * of this software must display the following
 * acknowledgment: "This product includes software
 * developed by Jesse Davis and Pedro in the 
 * Department of Computer Science and Engineering 
 * at the University of Washington".
 * 
 * 4. Your publications acknowledge the use or
 * contribution made by the Software to your research
 * using the following citation(s): 
 * Jesse Davis and Pedro Domingos, 
 * "Bottom-Up Learning of Markov Network Structure",
 * 27th International Conference on Machine Learning, 2010
 *
 * 
 * 5. Neither the name of the University of Washington nor
 * the names of its contributors may be used to endorse or
 * promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON
 * AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY
 * OF WASHINGTON OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
#ifndef FLOGPSEUDOLIKELIHOOD_H_AUG_18_2005
#define FLOGPSEUDOLIKELIHOOD_H_AUG_18_2005

#include <fstream>
#include <iostream>
#include <cmath>
#include "array.h"
#include "hashint.h"
#include "rule.h"
#include "timer.h"
  // expl and logl aren't available in cygwin / windows, so use exp and log
#ifndef expl
# define expl exp
# define logl log
#endif

////////////////////////// helper data structures //////////////////////////////
struct IndexAndCount 
{
  IndexAndCount() : index(NULL), count(0) {}
  IndexAndCount(int* const & i, const double& c) : index(i), count(c) {}
  int* index;
  double count;
};


  //LogPseudolikelihood is used in weight learning which makes the closed-world
  //assumption, so there is no ground predicate whose truth value is UNKNOWN

class FastPseudoLogLikelihood{

 public:
 FastPseudoLogLikelihood() : priorMeans_(NULL), priorStdDevs_(NULL)
    {
      updatePerAtom = true; //false;
      jUnify = false;
      jDebug = false;
      removeIdx = -1;
      counts = NULL;
      spy = false;
      spyGround = 2;
      debug =false;
      fastPll = false;
      multExPos = -1;
      //modifiedPositions = new IntHashArray;
    }

  ~FastPseudoLogLikelihood(){
   

   

    dataIdx.deleteItemsAndClear();
   
    //include.deleteItemsAndClear();
    
    //Array<Array<Array<IndexAndCount*>*>*>* counts;
    
    for(int ix = 0; ix < counts->size(); ix++){
      Array<Array<IndexAndCount*>*>* tmp1 = (*counts)[ix];
      for(int jx = 0; jx < tmp1->size(); jx++){
	Array<IndexAndCount*>* tmp2 = (*tmp1)[jx];
	//tmp2->deleteItemsAndClear();
	for(int kx = 0; kx < tmp2->size(); kx++){
	  if ((*tmp2)[kx]){
	    (*tmp2)[kx] = NULL;
	 }
	}
      }
      tmp1->deleteItemsAndClear();
    }
    satisfied.deleteItemsAndClear();
    oneOffSat.deleteItemsAndClear();
    counts->deleteItemsAndClear();
    //tmpPerAtomPll->deleteItemsAndClear();
    perAtomPll->deleteItemsAndClear();
    delete perAtomPll;
    //delete tmpPerAtomPll;
    delete counts;
    delete ruleIds;
  }

  void deleteCounts(){
    if (counts){
      for(int ix = 0; ix < counts->size(); ix++){
	Array<Array<IndexAndCount*>*>* tmp = (*counts)[ix];
	for (int jx = 0; jx < tmp->size(); jx++){
	  if ((*tmp)[jx]){
	    Array<IndexAndCount*>* tmp1 = (*tmp)[jx];
	    if (tmp1){
	      tmp1->deleteItemsAndClear();
	    }
	  }
	}
	tmp->deleteItemsAndClear();
      }
    }
    counts->deleteItemsAndClear();
    delete counts;
  }

  void setFastPLL(bool val){
    fastPll = val;
  }

  long double getFastValueAndGradient(double* const & gradient, const double* const & wt, const int& arrSize, int idx){
    if (idx < 0){
      perPredPllCache.growToSize(counts->size(),0);

      long double score =  getValueAndGradient(gradient, wt, arrSize);
      //jUnify = false;
      return score;
    }
    else{

      tmpPerPredPllCache.growToSize(counts->size(),0);
      long double score = getValueAndGradient(gradient, wt, arrSize, idx);
      //jUnify = false;
      return score;
    }
  }
  
  void print(int p, int g){
    Array<Array<IndexAndCount*>*>* gndingsToClauseIndexesAndCounts = (*counts)[p];
    Array<IndexAndCount*>* cs = (*gndingsToClauseIndexesAndCounts)[g];
    for(int ix = 0; ix < cs->size(); ix++){
      cout << "Idx " << *((*cs)[ix]->index) << " ";
    }
    cout << endl;
  }

  long double getValueAndGradient(double* const & gradient, const double* const & wt, const int& arrSize, int idx)
  {
    //double wpll = 0;
    memset(gradient, 0, arrSize*sizeof(double));
 
    long double wpll = 0; // weighted pseudo-log-likelihood
    long double* perPredGrad = new long double[arrSize];

    for (int p = 0; p < counts->size(); p++) // for each predicate
    {

      int compute = 0;
      if (ignoreSet.contains(p)){
	compute = 1;
      }
      else if (addSet.contains(p)){
	compute = 2;
      }
      //commented out: even though pred does not appear in any clause, each
      //of its groundings contribute -ln(2) to the wpll
      //if ((*gndPredClauseIndexesAndCounts)[p] == NULL) continue;
      //if (compute > 0){
	
      long double perPredPll = 0;
      memset(perPredGrad, 0, arrSize*sizeof(long double));
      Array<Array<IndexAndCount*>*>* gndingsToClauseIndexesAndCounts = (*counts)[p];
      long double oldPll = tmpPerAtomSum[p];
      //long double newSum = 0;
      //if pred p appears in one or more clauses
      if (fastPll && compute == 0){
	perPredPll = perPredPllCache[p];

	tmpPerPredPllCache[p] = perPredPllCache[p];

      }
      else if ((*counts)[p] != NULL){
	//use all groundings of predicate
	int numGnds = gndingsToClauseIndexesAndCounts->size();
	//assert(numGnds == (*((*numGndings_)[domainIdx]))[p]);
	if (fastPll){
	  if (compute == 1){
	    perPredPll = removedPerPredPllCache[p];

	  }
	  else{
	    for(int modIdx = 0; modIdx < modifiedPositions->size() ; modIdx++){
	      int g = (*modifiedPositions)[modIdx];
	      
	      if (removeIdx > -1){
		computePerPredPllAndGrad((*gndingsToClauseIndexesAndCounts)[g], 
					 g, wt, perPredPll, perPredGrad, 
					 (removeIdx > -1), idx, 
					 NULL, p, (*exWt)[g]);
	      }
	      else{
		computePerPredPllAndGrad((*gndingsToClauseIndexesAndCounts)[g], 
					 g, wt,  perPredPll, perPredGrad, 
					 (removeIdx > -1), idx, 
					 satisfied[idx], p, (*exWt)[g]);
	      }
	  
	    }//end of modIdx
	    if (missingFastCovPos > -1){
	     
	      computePerPredPllAndGrad((*gndingsToClauseIndexesAndCounts)[missingFastCovPos], 
				       missingFastCovPos, wt, perPredPll, 
				       perPredGrad, (removeIdx > -1), idx, 
				       NULL, p, (*exWt)[missingFastCovPos]);
	    }
	    int varPosition = positionIndex->find(p);
	    IntHashArray* updateGround = (*oneOffIndexes)[varPosition];
	    for(int modIdx = 0; modIdx < updateGround->size() ; modIdx++){
	      int g = (*updateGround)[modIdx];
	      
	      if (removeIdx > -1){
		computePerPredPllAndGrad((*gndingsToClauseIndexesAndCounts)[g], 
					 g, wt, 
					 perPredPll, perPredGrad, 
					 (removeIdx > -1), idx, 
					 NULL, p, (*exWt)[g]);
	      }
	      else{
		computePerPredPllAndGrad((*gndingsToClauseIndexesAndCounts)[g], 
					 g, wt,perPredPll, perPredGrad, 
					 (removeIdx > -1), idx, 
					 oneOffSat[idx], p, (*exWt)[g]);
	      }
	    
	    }//end of modIdx== 59


	    long double jScore = perPredPllCache[p] + perPredPll - oldPll;
	    perPredPll = jScore;
	  }//end of compute == 2
	}
	else{
	  for (int g = 0; g < numGnds; g++){ // for each grounding
	    int varPosition = positionIndex->find(p);
	    IntHashArray* updateGround = (*oneOffIndexes)[varPosition];
	    if (compute == 0){
	      computePerPredPllAndGrad((*gndingsToClauseIndexesAndCounts)[g], g,
				       wt, perPredPll, perPredGrad,
				       p, (*exWt)[g]);
	    }
	    else if (compute == 1){
	      computePerPredPllAndGrad((*gndingsToClauseIndexesAndCounts)[g],
				       g, wt, perPredPll, perPredGrad, true, 
				       idx, NULL, p, (*exWt)[g]); 
	    }
	    else{
	      if (modifiedPositions->contains(g)){
		computePerPredPllAndGrad((*gndingsToClauseIndexesAndCounts)[g], 
					 g, wt,  perPredPll, perPredGrad, 
					 false, idx, satisfied[idx],
					 p, (*exWt)[g]);
	      }
	      else if (updateGround->contains(g)){
		computePerPredPllAndGrad((*gndingsToClauseIndexesAndCounts)[g], 
					 g, wt, perPredPll, perPredGrad, false,
					 idx, oneOffSat[idx], p, (*exWt)[g]);
	      }
	      else{
		computePerPredPllAndGrad((*gndingsToClauseIndexesAndCounts)[g], 
					 g, wt, perPredPll, perPredGrad,
					 false, idx, NULL, p, (*exWt)[g]);
	      }
	    }
	  }
	}//end of for g loop
	

	tmpPerPredPllCache[p] = perPredPll;

      }//end if != null
      else{
	//pred p does not appear in any clauses
	perPredPll = 0;
	//perPredPll = (*count)[p]->size() * -log(2);
	//perPredGrad entries are all zeroes
      }
        
      //negate the value and gradient here
      wpll -= perPredPll;
 
      if (fastPll){
	if (removeIdx > -1){
	  gradient[removeIdx] -= perPredGrad[removeIdx];
	}
	else{
	  gradient[idx] -= perPredGrad[idx];
	}
       }
      else{
	for (int i = 0; i < arrSize; i++) 
	  {
	    gradient[i] -= perPredGrad[i]; 
	  }
      }
    } // for each predicate

    delete [] perPredGrad;

    //printWtsGradsWPLL(wt, gradient, arrSize, wpll); //for testing

      // if there are prior penalties
    if (numMeans_ > 0)
    {
      //commented out: numMeans_ can be larger than arrSize,so just consider
      //               the first arrSize items in priorMeans & priorStdDevs
      //assert(numMeans_ == arrSize);
        
      // subtract the gaussian priors

      for (int i = 0; i < arrSize; i++)
	{
	  
	  //since at this point the value and gradient have been negated,
	  //add the priors
	  wpll += (wt[i]-priorMeans_[i])*(wt[i]-priorMeans_[i])/
	    (2*priorStdDevs_[i]*priorStdDevs_[i]);         
	  if (fastPll){
	    if (removeIdx >= 0){
	      if (i == removeIdx){
		gradient[i] += (wt[i]-priorMeans_[i])/
		  (priorStdDevs_[i]*priorStdDevs_[i]);
	      }
	    }
	    else{
	      if (i == idx){
		gradient[i] += (wt[i]-priorMeans_[i])/
		(priorStdDevs_[i]*priorStdDevs_[i]);
	      }
	    }
	  }
	  else{
	    gradient[i] += (wt[i]-priorMeans_[i])/
	      (priorStdDevs_[i]*priorStdDevs_[i]);
	  }
	}
    }

    return wpll; 
  }



  long double getValueAndGradient(double* const & gradient, const double* const & wt, const int& arrSize)//, int x, int y)
  {
    memset(gradient, 0, arrSize*sizeof(double));

    long double wpll = 0; // weighted pseudo-log-likelihood
    long double* perPredGrad = new long double[arrSize];

    int numPreds = counts->size();
 
    for (int p = 0; p < numPreds; p++) // for each predicate
    {
      //cout << "p = " << p << endl;
      //commented out: even though pred does not appear in any clause, each
      //of its groundings contribute -ln(2) to the wpll
      //if ((*gndPredClauseIndexesAndCounts)[p] == NULL) continue;

      long double perPredPll = 0;
      memset(perPredGrad, 0, arrSize*sizeof(long double));
      Array<Array<IndexAndCount*>*>* gndingsToClauseIndexesAndCounts = (*counts)[p];
        //if pred p appears in one or more clauses
      if ((*counts)[p] != NULL)
      {
	
	//use all groundings of predicate
	int numGnds = gndingsToClauseIndexesAndCounts->size();
	//assert(numGnds == (*((*numGndings_)[domainIdx]))[p]);
	for (int g = 0; g < numGnds; g++) // for each grounding
	  {
	    //double t1 = nt.time();
	    Array<IndexAndCount*>* feats = (*gndingsToClauseIndexesAndCounts)[g];
	    computePerPredPllAndGrad(feats, g, wt, 
				     perPredPll, perPredGrad, p, (*exWt)[g]);
	  }
	//calls += numGnds;
      }
      else{ 
	//pred p does not appear in any clauses
	perPredPll = 0;
        //perPredGrad entries are all zeroes
      }

      //negate the value and gradient here
      perPredPllCache[p] = perPredPll;
      wpll -= perPredPll;
      if (debug)
	cout << "Per Pred Pll " << p << ": " << perPredPll << endl;

      for (int i = 0; i < arrSize; i++){
	  //(*predPGrad)[i] = perPredGrad[i];
          gradient[i] -= perPredGrad[i]; 
      }
    } // for each predicate

    delete [] perPredGrad;

    //printWtsGradsWPLL(wt, gradient, arrSize, wpll); //for testing

      // if there are prior penalties
    if (numMeans_ > 0)
    {
      //commented out: numMeans_ can be larger than arrSize,so just consider
      //               the first arrSize items in priorMeans & priorStdDevs
      //cerr << numMeans_ << " " << arrSize << endl;
      assert(numMeans_ >= arrSize);
      //cout << numMeans_ << " arrSize " << arrSize << endl;
      // subtract the gaussian priors
      for (int i = 0; i < arrSize; i++)
      {

        //since at this point the value and gradient have been negated,
        //add the priors
        wpll += (wt[i]-priorMeans_[i])*(wt[i]-priorMeans_[i])/
          (2*priorStdDevs_[i]*priorStdDevs_[i]);         
  
        gradient[i] += (wt[i]-priorMeans_[i])/
          (priorStdDevs_[i]*priorStdDevs_[i]);

      }
    }
    
    return wpll; 
  }
  void setJUnify(bool j){
    jUnify = j;
  }
  void setJ(bool j){
    jDebug = j;
  }

  void computePerPredPllAndGrad(const Array<IndexAndCount*>*
                                const& clauseIndexesAndCounts,
                                const int& g, const double* const & wt, 
                                long double& perPredPll, 
                                long double * const & perPredGrad, 
				int atomId,
				double exampleWt)
  {
      // compute W.(N^bar-N)

      // Changed to logl, expl because otherwise changing
      // order of .mln files changes weight results
    long double wdotn = 0;
    long double lpmb = 0;
    int dec = -1;
    long double iwdotn = 0;
    long double ilpmb = 0;
    long double decWt = 0;
    //assert(noRepeatedIndex(clauseIndexesAndCounts));
    //reset wdotn, because theise is for all grounds not just one grounding!!
    int numClausesUnifyWith = clauseIndexesAndCounts->size();

    for (int i = 0; i < numClausesUnifyWith; i++){
	// grounding g unifies with clause (*clauseIndexesAndCounts)[i].idx
        wdotn += wt[*( (*clauseIndexesAndCounts)[i]->index)] * (*clauseIndexesAndCounts)[i]->count;
	if ((*ruleIds)[(*( (*clauseIndexesAndCounts)[i]->index))] == g){
	  dec = i;
	  decWt = wt[*( (*clauseIndexesAndCounts)[i]->index)] * (*clauseIndexesAndCounts)[i]->count;
	}
    }
 
    if (wdotn > 0) lpmb +=  -logl(1+expl(-wdotn))-wdotn;
    else lpmb += -logl(1+expl(wdotn));

    if (dec > -1){
      iwdotn = wdotn - decWt;
      if (iwdotn > 0){
	ilpmb +=  -logl(1+expl(-iwdotn))-iwdotn;
      }
      else{
	ilpmb += -logl(1+expl(iwdotn));
      }
      (*((*perAtomPll)[atomId]))[g] = (((exampleWt-1) * lpmb) + ilpmb);
      perPredPll += (((exampleWt-1) * lpmb) + ilpmb);
    
    }
    else{
      (*((*perAtomPll)[atomId]))[g] = exampleWt * lpmb;
      perPredPll += (exampleWt * lpmb);
    }    
  

    // update gradient
    for (int i = 0; i < numClausesUnifyWith; i++){
      if (i == dec){
	perPredGrad[ *( (*clauseIndexesAndCounts)[i]->index ) ] 
	  += (((expl(lpmb)-1) * (*clauseIndexesAndCounts)[i]->count)* (exampleWt-1));
	
      }
      else if (dec > -1){
	
	perPredGrad[ *( (*clauseIndexesAndCounts)[i]->index ) ] 
	  += (((expl(lpmb)-1) * (*clauseIndexesAndCounts)[i]->count)* (exampleWt-1));
	perPredGrad[ *( (*clauseIndexesAndCounts)[i]->index ) ] 
	  += (((expl(ilpmb)-1) * (*clauseIndexesAndCounts)[i]->count));
	
      }
      else{
	perPredGrad[ *( (*clauseIndexesAndCounts)[i]->index ) ] 
	  += (((expl(lpmb)-1) * (*clauseIndexesAndCounts)[i]->count)* exampleWt);
	
      }
    }
  
  }

  void setExWt(Array<double>* exWts, Array<Array<int>*>* coverList){
    exWt = exWts;
    clist = coverList;
  }

  void initializeCounts(Array<Array<int>*>* data, Array<Rule*>* rules, Array<int*>* clauseIdx){

    if (counts){
      deleteCounts();
    }
    ruleIds = new Array<int>;
    ruleIds->growToSize(rules->size(),-1);
    
    satisfied.growToSize(rules->size(),NULL);
    oneOffSat.growToSize(rules->size(),NULL);
    //variables x datapoints x features
    counts = new Array<Array<Array<IndexAndCount*>*>*>;
    perAtomPll = new Array<Array<long double>*>;
    //tmpPerAtomPll = new Array<Array<long double>*>;
    int numFeats = (*data)[0]->size();
    tmpPerAtomSum.growToSize(numFeats,0);
    for(int ix = 0; ix < numFeats; ix++){
      Array<Array<IndexAndCount*>*>* tmp = new Array<Array<IndexAndCount*>*>;
      for(int jx =0; jx < data->size(); jx++){
	Array<IndexAndCount*>* tmp1 = new Array<IndexAndCount*>;
	tmp->append(tmp1);
      }
      Array<long double>* tmpD1 = new Array<long double>;
      tmpD1->growToSize(data->size(),0);
      perAtomPll->append(tmpD1);

      counts->append(tmp);
      if (ix == 0){
	cout << "perAtomPll = " << numFeats << " x ";
	cout << tmpD1->size() << endl;
      }
    }

    for(int ix = 0; ix < rules->size(); ix++){

      (*ruleIds)[ix] = (*rules)[ix]->getRid();
      //cout << "ix " << ix << " " << *((*clauseIdx)[ix]) << " " << (*ruleIds)[ix] << endl;
      IndexAndCount* sat = new IndexAndCount((*clauseIdx)[ix],-1);
      IndexAndCount* oneOff = new IndexAndCount((*clauseIdx)[ix],1);
      satisfied[ix] = sat;
      oneOffSat[ix] = oneOff;
      oneOffIndexes = (*rules)[ix]->oneOffPos;
      for(int jx = 0; jx < data->size(); jx++){
	Array<int>* sameEx = (*clist)[jx];

	if (jx == (*rules)[ix]->getRid() && sameEx->size() == 1){
	  for(int cx = 0; cx < sameEx->size(); cx++){
	    (*rules)[ix]->covered->append((*sameEx)[cx]);
	  }
	}
	else if (jx == (*rules)[ix]->getRid()){
	  
	  (*rules)[ix]->fastCov->append(jx);
	  for(int cx = 0; cx < sameEx->size(); cx++){
	    (*rules)[ix]->covered->append((*sameEx)[cx]);
	  }
	  IntHashArray* vars = (*rules)[ix]->index;
	  for(int kx = 0; kx < vars->size(); kx++){
	    int varId = (*vars)[kx];
	    (*((*counts)[varId]))[jx]->append(sat);
	  }
	}
	else if ((*rules)[ix]->covers((*data)[jx])){
	  (*rules)[ix]->fastCov->append(jx);
	  for(int cx = 0; cx < sameEx->size(); cx++){
	    (*rules)[ix]->covered->append((*sameEx)[cx]);
	  }
	  IntHashArray* vars = (*rules)[ix]->index;
	  for(int kx = 0; kx < vars->size(); kx++){
	    int varId = (*vars)[kx];
	    (*((*counts)[varId]))[jx]->append(sat);
	  }
	}
	else{
	  IntHashArray* vars = (*rules)[ix]->index;
	  for(int kx = 0; kx < vars->size(); kx++){
	    int varId = (*vars)[kx];
	    int tmpVal = (*((*data)[jx]))[varId];
	    (*((*data)[jx]))[varId] = (tmpVal + 1) % 2;
	    if ((*rules)[ix]->covers((*data)[jx])){
	      (*((*counts)[varId]))[jx]->append(oneOff);
	      (*oneOffIndexes)[kx]->append(jx);
	    }
	    (*((*data)[jx]))[varId] = tmpVal;
	  }
	}
      }
    }
    bool printInit = false;
    if (printInit){
      for(int ix = 0; ix < counts->size(); ix++){
	Array<Array<IndexAndCount*>*>* tmp1 = (*counts)[ix];
	for(int jx = 0; jx < tmp1->size(); jx++){
	  Array<IndexAndCount*>* tmp2 = (*tmp1)[jx];
	  cout << "ix " << ix << " jx " << jx << " : ";
	  for(int kx = 0; kx < tmp2->size(); kx++){
	    cout << "(" << *((*tmp2)[kx]->index) << "," << (*tmp2)[kx]->count << ") ";
	  }
	  cout << endl;
	}
	
      }
    }
  }


  void updateCache(){
    if (perPredPllCache.size() != tmpPerPredPllCache.size()){
      perPredPllCache.growToSize(tmpPerPredPllCache.size(), 0);
    }
    for(int ix = 0; ix < tmpPerPredPllCache.size(); ix++){
      perPredPllCache[ix] = tmpPerPredPllCache[ix];
    }
  }

  void updateModel(bool remove, int ruleIdx, int ruleId){
    (*ruleIds)[ruleIdx] = ruleId;
    for(int ix = 0; ix < ignoreSet.size(); ix++){
      int p = ignoreSet[ix];
 
      for(int modIdx = 0; modIdx < oldModifiedPositions->size() ; modIdx++){
	int g = (*oldModifiedPositions)[modIdx];
	Array<IndexAndCount*>* tmp = (*((*counts)[p]))[g];
	for(int kx = 0; kx < tmp->size(); kx++){
	  if (*((*tmp)[kx]->index) == ruleIdx){
	    
	    tmp->removeItemFastDisorder(kx);
	  }
	}	
      }
      int varPosition = oldPositionIndex->find(p);
      IntHashArray* updateGround = (*oldOneOffIndexes)[varPosition];
      for(int modIdx = 0; modIdx < updateGround->size() ; modIdx++){
	int g = (*updateGround)[modIdx];
	Array<IndexAndCount*>* tmp = (*((*counts)[p]))[g];
	for(int kx = 0; kx < tmp->size(); kx++){
	  if (*((*tmp)[kx]->index) == ruleIdx){
	    
	    tmp->removeItemFastDisorder(kx);
	  }
	}
      }
    }
    
    if (!remove){
      for(int ix = 0; ix < addSet.size(); ix++){
	int p = addSet[ix];
	Array<Array<IndexAndCount*>*>* data = (*counts)[p];
	//cerr << "HERE - iter " << ix << "\n";
	for(int modIdx = 0; modIdx < modifiedPositions->size() ; modIdx++){
	  int g = (*modifiedPositions)[modIdx];
	  bool found = false;
	  Array<IndexAndCount*>* clauses = (*data)[g];
	  for(int kx = 0; kx < clauses->size() && !found; kx++){
	    if (*((*clauses)[kx]->index) == ruleIdx){
	      //FLAG - Delete Problem
	      found = true;
	      (*clauses)[kx] = satisfied[ruleIdx];
	    }
	  }
	  if (!found){
	    clauses->append(satisfied[ruleIdx]);
	  }
	}
	int varPosition = positionIndex->find(p);
	IntHashArray* updateGround = (*oneOffIndexes)[varPosition];
	for(int modIdx = 0; modIdx < updateGround->size() ; modIdx++){
	  int g = (*updateGround)[modIdx];
	  bool found = false;
	  Array<IndexAndCount*>* clauses = (*data)[g];
	  for(int kx = 0; kx < clauses->size() && !found; kx++){
	    if (*((*clauses)[kx]->index) == ruleIdx){
	      //FLAG - Delete Problem
	      found = true;
	      (*clauses)[kx] = oneOffSat[ruleIdx];
	    }
	  }
	  if (!found){
	    clauses->append(oneOffSat[ruleIdx]);
	  }
	}
      }
    }
    else{
      for(int ix = 0; ix < addSet.size(); ix++){
	int p = addSet[ix];

	for(int modIdx = 0; modIdx < modifiedPositions->size() ; modIdx++){
	  int g = (*modifiedPositions)[modIdx];
	  Array<IndexAndCount*>* tmp = (*((*counts)[p]))[g];
	  for(int kx = 0; kx < tmp->size(); kx++){
	    if (*((*tmp)[kx]->index) == ruleIdx){
	      IndexAndCount* del = tmp->removeItemFastDisorder(kx);
	    }
	  }	
	}
	if (missingFastCovPos > -1){
	  Array<IndexAndCount*>* tmp = (*((*counts)[p]))[missingFastCovPos];
	  for(int kx = 0; kx < tmp->size(); kx++){
	    if (*((*tmp)[kx]->index) == ruleIdx){
	      IndexAndCount* del = tmp->removeItemFastDisorder(kx);
	    }
	  }	
	}
	int varPosition = positionIndex->find(p);
	IntHashArray* updateGround = (*oneOffIndexes)[varPosition];
	for(int modIdx = 0; modIdx < updateGround->size() ; modIdx++){
	  int g = (*updateGround)[modIdx];
	  Array<IndexAndCount*>* tmp = (*((*counts)[p]))[g];
	  for(int kx = 0; kx < tmp->size(); kx++){
	    if (*((*tmp)[kx]->index) == ruleIdx){
	      IndexAndCount* del = tmp->removeItemFastDisorder(kx);
	    }
	  }
	}
      }
    }
  }

  void removeUpdate(int rulePos){
    ruleIds->removeItemFastDisorder(rulePos);
    IndexAndCount* del1 = satisfied.removeItemFastDisorder(rulePos);
    IndexAndCount* del2 = oneOffSat.removeItemFastDisorder(rulePos);
    delete del1;
    delete del2;
  }

  void countNewRule(Rule* rule, Rule* oldRule, int* ruleIdx, Array<Array<int>*>* data, IntHashArray* ignore, bool updatePerAtomCache){
    missingFastCovPos = -1;//newRule->getRid();
    multExPos = -1;
    removeIdx = -1;
    includeRuleId = rule->getRid();
    updatePerAtom = updatePerAtomCache;
    oneOffIndexes = rule->oneOffPos;
    positionIndex = rule->index;
    oldPositionIndex = oldRule->index;
    oldOneOffIndexes = oldRule->oneOffPos;
    oldModifiedPositions = oldRule->fastCov;
    IntHashArray* nvars = rule->index;
    addSet.clear();
    ignoreSet.clear();
    for(int ix = 0; ix < ignore->size(); ix++){
      ignoreSet.append((*ignore)[ix]);
    }
    for(int ix = 0; ix < nvars->size(); ix++){
      addSet.append((*nvars)[ix]);
    }

    for(int jx = 0; jx < data->size(); jx++){
      Array<int>* sameEx = (*clist)[jx];
      if (jx == rule->getRid() && sameEx->size() == 1 ){
	for(int cx = 0; cx < sameEx->size(); cx++){
	  rule->covered->append((*sameEx)[cx]);
	}
      }
      else if (jx == rule->getRid()){
	multExPos = jx;
	rule->fastCov->append(jx);
	for(int cx = 0; cx < sameEx->size(); cx++){
	  rule->covered->append((*sameEx)[cx]);
	}
      }
      else if (rule->covers((*data)[jx])){
	rule->fastCov->append(jx);
	for(int cx = 0; cx < sameEx->size(); cx++){
	  rule->covered->append((*sameEx)[cx]);
	}
      }
      else{
	IntHashArray* vars = rule->index;
	for(int kx = 0; kx < vars->size(); kx++){
	  int varId = (*vars)[kx];
	  int tmpVal = (*((*data)[jx]))[varId];
	  (*((*data)[jx]))[varId] = (tmpVal + 1) % 2;
	  if (rule->covers((*data)[jx])){
	    (*oneOffIndexes)[kx]->append(jx);
	  }
	  (*((*data)[jx]))[varId] = tmpVal;
	}
      }
    }
    modifiedPositions = rule->fastCov;
    sum();
  }

  void computePerPredPllAndGrad(const Array<IndexAndCount*>*
                                const& clauseIndexesAndCounts,
                                const int& g, const double* const & wt, 
                                long double& perPredPll, 
                                long double * const & perPredGrad,
				bool ignore,
				int idx,
				IndexAndCount* include,
				int atomId,
				double exampleWt)
  {
      // compute W.(N^bar-N)

      // Changed to logl, expl because otherwise changing
      // order of .mln files changes weight results
    long double wdotn = 0;
    long double lpmb = 0;
    bool found = false; 
    int dec = -2;
    long double iwdotn = 0;
    long double ilpmb = 0;
    long double decWt = 0;
    int numClausesUnifyWith = clauseIndexesAndCounts->size();
    
    for (int i = 0; i < numClausesUnifyWith; i++){     
      int tmpIdx =  *( (*clauseIndexesAndCounts)[i]->index );

      if (ignore && idx == tmpIdx){
	//clause no longer applies
      }
      else if (tmpIdx == idx){
	if ((*ruleIds)[tmpIdx] == g){
	  dec = i;
	  decWt = wt[ *(include->index ) ] * include->count;
	}
	found = true;
	wdotn += wt[ *(include->index ) ] * include->count;

      }
      else{
	// grounding g unifies with clause (*clauseIndexesAndCounts)[i].idx
	wdotn += wt[ *( (*clauseIndexesAndCounts)[i]->index ) ] * 
	  (*clauseIndexesAndCounts)[i]->count;	
	if ((*ruleIds)[tmpIdx] == g){
	  dec = i;
	  decWt = wt[ *( (*clauseIndexesAndCounts)[i]->index ) ] * 
	    (*clauseIndexesAndCounts)[i]->count;
	}

      }
    }
  
    if (!found && include){
      wdotn += wt[ *(include->index ) ] * include->count;
      if ( includeRuleId == g){
	dec = -1;
	decWt = wt[ *(include->index ) ] * include->count;
      }

    }
 
    if (wdotn > 0) lpmb += -logl(1+expl(-wdotn))-wdotn;
    else lpmb += -logl(1+expl(wdotn));
    
    if (dec > -2){
      iwdotn = wdotn - decWt;
      if (iwdotn > 0){
	ilpmb +=  -logl(1+expl(-iwdotn))-iwdotn;
      }
      else{
	ilpmb += -logl(1+expl(iwdotn));
      }
      if (updatePerAtom){
	(*((*perAtomPll)[atomId]))[g] = (((exampleWt-1) * lpmb) + ilpmb);
      }
      perPredPll += (((exampleWt-1) * lpmb) + ilpmb);
     
    }
    else{
      if (updatePerAtom){
	(*((*perAtomPll)[atomId]))[g] = exampleWt * lpmb;
      }
      perPredPll += (exampleWt * lpmb);
  
    }
   
    for (int i = 0; i < clauseIndexesAndCounts->size(); i++){
      int tmpIdx =  *( (*clauseIndexesAndCounts)[i]->index );
      if (ignore && tmpIdx == idx){
	//do nothing :)
      }
      else if (tmpIdx == idx){
	//drop one of the examples
	if (i == dec){

	  perPredGrad[ *( include->index ) ] 
	    += ((exampleWt-1) * (expl(lpmb)-1) * include->count);

	}
	//really two examples, so treat seperately
	else if (dec > -2){

	  perPredGrad[ *( include->index ) ] 
	    += ((exampleWt-1) * (expl(lpmb)-1) * include->count);
	  perPredGrad[ *( include->index ) ] 
	    += ((expl(ilpmb)-1) * include->count);	
	}
	else{

	  perPredGrad[ *( include->index ) ] 
	    += (exampleWt * (expl(lpmb)-1) * include->count);

	}
      }
      else if (removeIdx == tmpIdx){
	//this feature doesn't apply to one of the examples
	if (i == dec){
	  perPredGrad[ *( (*clauseIndexesAndCounts)[i]->index ) ] 
	    += (((expl(lpmb)-1) * (*clauseIndexesAndCounts)[i]->count)* (exampleWt-1));
	}
	//really two examples, so treat seperately
	else if (dec > -2){
	  perPredGrad[ *( (*clauseIndexesAndCounts)[i]->index ) ] 
	    += (((expl(lpmb)-1) * (*clauseIndexesAndCounts)[i]->count)* (exampleWt-1));
	  perPredGrad[ *( (*clauseIndexesAndCounts)[i]->index ) ] 
	    += (((expl(ilpmb)-1) * (*clauseIndexesAndCounts)[i]->count));
	}
	//this feature applies to all examples
	else{
	  perPredGrad[ *( (*clauseIndexesAndCounts)[i]->index ) ] 
	    += (((expl(lpmb)-1) * (*clauseIndexesAndCounts)[i]->count)* exampleWt);
	}
      }
    }
    
    if (!found && include){
      //really two examples and one example deo treat seperately
      if (dec == -1){
	perPredGrad[ *( include->index ) ] 
	  += ((exampleWt-1) * (expl(lpmb)-1) * include->count);
      }
      //really two examples, so treat seperately
      else if (dec > -2){
	perPredGrad[ *( include->index ) ] 
	  += ((exampleWt-1) * (expl(lpmb)-1) * include->count);
	perPredGrad[ *( include->index ) ] 
	  += ((expl(ilpmb)-1) * include->count);	
      }
      else{
	perPredGrad[ *( include->index ) ] 
	  += (exampleWt * (expl(lpmb)-1) * include->count);
      }
      /*
	perPredGrad[ *( include->index ) ] 
	+= (exampleWt * ((expl(lpmb)-1) * include->count));
      */
    }

  }

  void setRemovedPll(const double* const & wt, int idx){
    removedPerPredPllCache.growToSize(counts->size(),0);
    for(int removed = 0; removed < ignoreSet.size(); removed++){
      int p = ignoreSet[removed];
      Array<Array<IndexAndCount*>*>* gndingsToClauseIndexesAndCounts = (*counts)[p];
      long double perPredPll = 0;
      long double oldPll = 0;
      for(int modIdx = 0; modIdx < oldModifiedPositions->size() ; modIdx++){
	int g = (*oldModifiedPositions)[modIdx];
	oldPll+= (*((*perAtomPll)[p]))[g];
	computePerPredPll((*gndingsToClauseIndexesAndCounts)[g], g, wt, perPredPll, idx, p, (*exWt)[g]);
      }
      int varPosition = oldPositionIndex->find(p);

      IntHashArray* updateGround = (*oldOneOffIndexes)[varPosition];
      for(int modIdx = 0; modIdx < updateGround->size() ; modIdx++){
	int g = (*updateGround)[modIdx];
	oldPll+= (*((*perAtomPll)[p]))[g];
	computePerPredPll((*gndingsToClauseIndexesAndCounts)[g], g, wt, perPredPll, idx, p, (*exWt)[g]);
      }
      double tmpPllScore = perPredPllCache[p] + perPredPll - oldPll;
      removedPerPredPllCache[p] = tmpPllScore;
    }
  }

  void computePerPredPll(const Array<IndexAndCount*>*
			 const& clauseIndexesAndCounts,
			 const int& g, const double* const & wt, 
			 long double& perPredPll, 
			 int idx,
			 int atomId,
			 double exampleWt)
  {
      // compute W.(N^bar-N)

      // Changed to logl, expl because otherwise changing
      // order of .mln files changes weight results
    long double wdotn = 0;
    long double lpmb = 0;
    int numClausesUnifyWith = clauseIndexesAndCounts->size();
    int dec = -1;
    long double decWt = 0;
    long double iwdotn = 0;
    long double ilpmb = 0;
 
    for (int i = 0; i < numClausesUnifyWith; i++){
 
      int tmpIdx =  *( (*clauseIndexesAndCounts)[i]->index );

      if (idx != tmpIdx){
	// grounding g unifies with clause (*clauseIndexesAndCounts)[i].idx
	wdotn += wt[ *( (*clauseIndexesAndCounts)[i]->index ) ] * 
	  (*clauseIndexesAndCounts)[i]->count;
	if ((*ruleIds)[(*( (*clauseIndexesAndCounts)[i]->index))] == g){
	  dec = i;
	  decWt = wt[*( (*clauseIndexesAndCounts)[i]->index)] * (*clauseIndexesAndCounts)[i]->count;
	}

      }
    }

    if (wdotn > 0) lpmb += -logl(1+expl(-wdotn))-wdotn;
    else lpmb += -logl(1+expl(wdotn));
    
    if (dec > -1){
      iwdotn = wdotn - decWt;
      if (iwdotn > 0){
	ilpmb +=  -logl(1+expl(-iwdotn))-iwdotn;
      }
      else{
	ilpmb += -logl(1+expl(iwdotn));
      }
      if (updatePerAtom){
	(*((*perAtomPll)[atomId]))[g] = (((exampleWt-1) * lpmb) + ilpmb);
      }
      perPredPll += (((exampleWt-1) * lpmb) + ilpmb);

    }
    else{
      if (updatePerAtom){
	(*((*perAtomPll)[atomId]))[g] = exampleWt * lpmb;
      }
      perPredPll += (exampleWt * lpmb);
 
    }
    /*
      if (updatePerAtom)
      (*((*perAtomPll)[atomId]))[g] = exampleWt * lpmb;
      perPredPll += (exampleWt * lpmb);
    */
  }

    //set numMeans to -1 if there is no prior
  void setMeansStdDevs(const int& arrSize, const double* const & priorMeans, 
                       const double* const & priorStdDevs)
  {
    numMeans_ = arrSize;
    priorMeans_ = priorMeans;
    priorStdDevs_ = priorStdDevs;
  }

  void sum(){
    
    for(int ix = 0; ix < addSet.size(); ix++){
      int p = addSet[ix];
      long double oldPll = 0;
      for(int modIdx = 0; modIdx < modifiedPositions->size() ; modIdx++){
	int g = (*modifiedPositions)[modIdx];
	oldPll+= (*((*perAtomPll)[p]))[g];
      }
      if (missingFastCovPos > -1){
       	oldPll+= (*((*perAtomPll)[p]))[missingFastCovPos];
      }
      int varPosition = positionIndex->find(p);
      IntHashArray* updateGround = (*oneOffIndexes)[varPosition];
      for(int modIdx = 0; modIdx < updateGround->size() ; modIdx++){
	int g = (*updateGround)[modIdx];
	oldPll+= (*((*perAtomPll)[p]))[g];
      }
      
      tmpPerAtomSum[p] = oldPll;
    }
  }


  void setIgnoreSet(IntHashArray* iset, Rule* oldRule, Rule* newRule, int dupRuleIdx, bool updatePerAtomCache){
    updatePerAtom = updatePerAtomCache;
    removeIdx = dupRuleIdx;
    includeRuleId = -1;
    ignoreSet.clear();
    addSet.clear();
    oldModifiedPositions = oldRule->fastCov;
    oldOneOffIndexes = oldRule->oneOffPos;
    oldPositionIndex = oldRule->index;
    modifiedPositions = newRule->fastCov;
    oneOffIndexes = newRule->oneOffPos;
    positionIndex = newRule->index;
    for(int ix = 0; ix < iset->size(); ix++){
      ignoreSet.append((*iset)[ix]);
    }
    if (oldRule->fastCov->contains(newRule->getRid()) &&
	!newRule->fastCov->contains(newRule->getRid())){
      missingFastCovPos = newRule->getRid();
    }
    else{
      missingFastCovPos = -1;//newRule->getRid();
    }
    IntHashArray* aset = newRule->index;
    for(int ix = 0; ix < aset->size(); ix++){
      addSet.append((*aset)[ix]);
    }
    sum();
  }

 private:
  bool jUnify;
  bool jDebug;
  bool fastPll;
  bool debug;
  bool spy;
  bool updatePerAtom;
  int missingFastCovPos;
  int spyGround;
  int numMeans_;
  const double* priorMeans_;
  const double* priorStdDevs_;
  Array<IndexAndCount*> satisfied;
  Array<IndexAndCount*> oneOffSat;
  Array<int>* ruleIds;
  
  int multExPos;
  Array<long double> perPredPllCache;
  Array<long double> tmpPerPredPllCache;
  Array<long double> removedPerPredPllCache;

  Array<Array<Array<IndexAndCount*>*>*>* counts;
  int removeIdx;
  int includeRuleId;
  IntHashArray ignoreSet;
  IntHashArray addSet;
  Array<int> predIdx;
  Array<Array<int>*> dataIdx;

  IntHashArray* modifiedPositions;
  IntHashArray* positionIndex;
  IntHashArray* oldModifiedPositions;
  IntHashArray* oldPositionIndex;
  Array<IntHashArray*>* oneOffIndexes;
  Array<IntHashArray*>* oldOneOffIndexes;

  Array<Array<long double>*>* perAtomPll;
  Array<long double> tmpPerAtomSum;

  Array<double>* exWt;
  Array<Array<int>*>* clist;
};
#endif
