/*
 * All of the documentation and software included in the
 * BLM Software is copyrighted by Jesse Davis and Pedro Domingos.
 * The arguments.cpp, arguments.h, hasharray.h, hash.h, hashint.h
 * flbfgsb.h, lbfgsb.h, random.h and timer.h come from the
 * Alchemy Software system. These files are denoted as coming from
 * Alchemy.
 * 
 * Copyright [2010] Jesse Davis and Pedro Domingos All rights reserved.
 * 
 * Contact: Jesse Davis, University of Washington
 * (jdavis@cs.washington.edu).
 * 
 * Redistribution and use in source and binary forms, with
 * or without modification, are permitted provided that
 * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above
 * copyright notice, this list of conditions and the
 * following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the
 * above copyright notice, this list of conditions and the
 * following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use
 * of this software must display the following
 * acknowledgment: "This product includes software
 * developed by Jesse Davis and Pedro in the 
 * Department of Computer Science and Engineering 
 * at the University of Washington".
 * 
 * 4. Your publications acknowledge the use or
 * contribution made by the Software to your research
 * using the following citation(s): 
 * Jesse Davis and Pedro Domingos, 
 * "Bottom-Up Learning of Markov Network Structure",
 * 27th International Conference on Machine Learning, 2010
 *
 * 
 * 5. Neither the name of the University of Washington nor
 * the names of its contributors may be used to endorse or
 * promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON
 * AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY
 * OF WASHINGTON OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
#ifndef RULE_H_NOV_23_2005
#define RULE_H_NOV_23_2005

#include "array.h"
#include "hashint.h"

struct Rule{
  Rule(int rid){
    id = rid;
    value = new Array<int>;
    index = new IntHashArray;
    covered = new IntHashArray;
    fastCov = new IntHashArray;
    oneOffPos = new Array<IntHashArray*>;
  }
  
  
  
Rule(IntHashArray* idx, Array<int>* val) : index(idx), value(val) 
  {
    covered = new IntHashArray;
    
    fastCov = new IntHashArray;
    oneOffPos = new Array<IntHashArray*>;
    for(int ix = 0; ix < index->size(); ix++){
      IntHashArray* tmp = new IntHashArray;
      oneOffPos->append(tmp);
    }
    id = -1;
  }
  ~Rule(){
    //cout << "Deleting " << id << endl;
    index->clear();
    covered->clear();
    
    value->clear();
    oneOffPos->deleteItemsAndClear();
    delete index;
    delete covered;
    delete value;
    delete fastCov;
    delete oneOffPos;
  }
  
  IntHashArray* fastCov;
  int id;
  IntHashArray* index;
  Array<int>* value;
  IntHashArray* covered;
  Array<IntHashArray*>* oneOffPos;

  void addValue(int val, int position){
    value->append(val);
    index->append(position);
        IntHashArray* tmp = new IntHashArray;
    oneOffPos->append(tmp);
  }
  bool covers(Array<int>* example){
    
    for(int ix = 0; ix < index->size(); ix++){
      
      int pos = (*index)[ix];
      if ((*value)[ix] != (*example)[pos]){
	return false;
      }
    }
    return true;
  }
  bool equals(Rule* nRule){
    IntHashArray* idx = nRule->index;  
    Array<int>* val = nRule->value;
    if (idx->size() != index->size()){
      return false;
    }
    for(int ix = 0; ix < index->size(); ix++){
      int pos = idx->find((*index)[ix]);
      if (pos < 0){
 	return false;
      }
      else if ((*val)[pos] != (*value)[pos]){
	return false;
      }
    }
    return true;
  }

  int ruleLength(){
    return index->size();
  }

  int getRid(){
    return id;
  }

  void setRid(int rid){
    id = rid;
  }
 ostream& printBrief(ostream& out) const
  {
    for(int ix = 0; ix < index->size(); ix++){
      out << (*index)[ix] << ":" << (*value)[ix] << " ";
    }
    out << "(";
    for(int ix = 0; ix < fastCov->size(); ix++){
      out << (*fastCov)[ix];
      if (ix < (fastCov->size()-1)){
	out << " ";
      }
    }    
    out << ") RID " << id;
    return out;
  }

  ostream& print(ostream& out) const
  {
    for(int ix = 0; ix < index->size(); ix++){
      out << (*index)[ix] << ":" << (*value)[ix] << " ";
    }
    out << "(";
    for(int ix = 0; ix < covered->size(); ix++){
      out << (*covered)[ix];
      if (ix < (covered->size()-1)){
	out << " ";
      }
    }
    out << "/";
    for(int ix = 0; ix < fastCov->size(); ix++){
      out << (*fastCov)[ix];
      if (ix < (fastCov->size()-1)){
	out << " ";
      }
    }    
    out << ") " << covered->size() << " RID " << id;
    return out;
  }

  ostream& printMLN(ostream& out) const
  {
    for(int ix = 0; ix < index->size(); ix++){
      if ((*value)[ix] == 1){
	out << "P" << (*index)[ix] << "(id1)";// << (*value)[ix] << " ";
      }
      else{
	out << "!P" << (*index)[ix] << "(id1)";// << (*value)[ix] << " ";
      }
      if (ix < index->size()-1){
	out << " ^ ";
      }
    }
    return out;
  }
};

#endif
